import {
  BANK_INITIAL_REFRESH_TOKEN,
  BANK_OAUTH_CLIENT_ID,
  BANK_OAUTH_CLIENT_SECRET,
} from "@/config.ts"
import logger from "@common/logger.ts"
import dayjs, { type Dayjs } from "dayjs"
import type { Database } from "better-sqlite3"
import {
  clearTokens,
  fetchToken,
  insertTokens,
  type TokenResponse,
} from "@/bank/db/queries.ts"
import * as Oauth from "@sb1/oauth.ts"
import * as Transactions from "@sb1/transactions.ts"
import type { ActualTransaction } from "@/actual.ts"
import { bankTransactionIntoActualTransaction } from "@/mappings.ts"
import type { OAuthTokenResponse } from "@sb1/types.ts"

export interface Bank {
  fetchTransactions: (
    interval: Interval,
    ...accountKeys: ReadonlyArray<string>
  ) => Promise<ReadonlyArray<ActualTransaction>>
}

export interface Interval {
  fromDate: Dayjs
  toDate: Dayjs
}

export class Sparebank1Impl implements Bank {
  private readonly db: Database

  constructor(db: Database) {
    this.db = db
  }

  async fetchTransactions(
    interval: Interval,
    ...accountKeys: ReadonlyArray<string>
  ): Promise<ReadonlyArray<ActualTransaction>> {
    const response = await Transactions.list(
      await this.getAccessToken(),
      accountKeys,
      interval,
    )
    const sparebankTransactions = response.transactions
    return sparebankTransactions.map(bankTransactionIntoActualTransaction)
  }

  private async getAccessToken(): Promise<string> {
    const accessToken = fetchToken(this.db, "access-token")

    if (accessToken && this.isValidToken(accessToken)) {
      return accessToken.token
    }
    const response = await this.fetchNewTokens()
    return response.access_token
  }

  private isValidToken(tokenResponse: TokenResponse): boolean {
    // TODO make sure the same timezone is used. Db uses UTC
    return dayjs().isBefore(tokenResponse.expires_at)
  }

  private async getRefreshToken(): Promise<string> {
    const tokenResponse = fetchToken(this.db, "refresh-token")

    if (!tokenResponse) {
      return BANK_INITIAL_REFRESH_TOKEN
    } else if (this.isValidToken(tokenResponse)) {
      return tokenResponse.token
    }
    logger.warn("Refresh token expired, deleting tokens from database")
    clearTokens(this.db)
    throw new Error("Refresh token is expired. Create a new one")
  }

  private async fetchNewTokens(): Promise<OAuthTokenResponse> {
    const refreshToken = await this.getRefreshToken()
    const result = await Oauth.refreshToken(
      BANK_OAUTH_CLIENT_ID,
      BANK_OAUTH_CLIENT_SECRET,
      refreshToken,
    )

    if (result.status === "failure") {
      throw new Error(`Failed to fetch refresh token: '${result.data}'`)
    }
    const oAuthToken = result.data

    insertTokens(this.db, oAuthToken)
    return oAuthToken
  }
}
